/** Two recursive solutions.
  
  We presented method updateCounts in class on Tuesday, 17 February, in order to
  present some interesting string manipulation in the simplest manner.
   
  However, updateCounts uses VERY POOR STYLE, especially in its (ab)use of static variables. 
  It uses them as so-called "global variables", which are frowned upon by all in the know, 
  including us. We do not write programs like this in practice, and neither should you.
   
  To give an analogy: you could write someone a note consisting of the two words "You stink".
  This is short, it's very clear what you mean, and you might even feel good writing it.
  However, we guarantee that it will hurt you in the long run.
   
  In the code below, we present a different solution, which is not only in better style
  but is more efficient. The second solution is called using two functions called hitsAndMisses,
  one to set things up and the other to do the calculation. 
  
  We explain two major differences between the two implementations.
  
  1. updateCounts does everything using global variables.
  hitsAndMisses does everything using parameters and return values.
  
  Two values are supposed to be returned: the number of hits and the number of misses.
  But function hitsAndMisses can return only one thing. To solve this problem we use
  a class Pair, which wraps TWO ints, and return an instance of class Pair.
  
  2. Procedure updateCounts is very inefficient, and hitsAndMisses is much more efficient
  (it takes far less time than updateCounts).
  You can see this when you execute the following two calls in the interactions pane.
  
  LieDemoBarbaricVersion.produceCounts("09-powell.htm", "destruction", true)
  
  LieDemoBarbaricVersion.produceCounts("09-powell.htm", "destruction", false)
  
  Here is why. produceCounts continually extracts substrings and joins two strings (using +),
  and this takes time. To create a new string that is the + of two strings takes time
  proportional to the sum of the lengths of the strings. Thus, to catenate two strings of
  length 1000 and 2000 characters takes time proportional to 3000. Do this very often, and
  it takes a lot of time.
  
  Function hitsAndMisses doesn't construct new strings at all! It just keeps track of what it
  has looked at in the one string, text. So it doesn't do a lot of string copying.
  
  */

import java.lang.Math.*;

/** Class houses methods dealing with testing hypotheses about deceptive language.
  * Uses class In from http://www.cs.princeton.edu/introcs/stdlib/In.java.html
  * (save it in the directory that contains this program)
  */
public class LieDemoBarbaricVersion {
    /** Strings that define the start and end of a span. */
    public static final String spanStart= "<font color=\"red\">";
    public static final String spanEnd= "</font>";
    
    /** computed outside this program, but useful info for printouts. */
    /** provide ratio info in output if true */
    public static boolean printRatioInfo= false; 
    
    /** computed through other means */
    public static final double unigramHitMissRatio= 0.314328; 
    
    
    /* w: target word (sequence of lowercase letters, no blanks or punctuation)
     hits and misses: number of hits (occurrences within a span) and 
     misses (occurrences outside a span) known so far for w.
     text: has yet to be processed for occurrences. These are the "global
     variables", used in several methods, which constitute a bad practice. */
    private static String w;
    private static int hits;   
    private static int misses;  
    private static String text; 
    
    
    /**= String of the form "Hits: h; Misses m" (plus external info depending on  
      printRatioInfo), where for character sequence target,
      h = no. of times target occurs within a span  in file sourceName
      m = no. of times target occurs outside a span in sourceName
      
      If choice is true, use procedure updateCounts() to count the number of hits
      and misses; otherwise, use function hitsAndMisses.
      
      Preconditions:
      target consists only of letters, and at least one, all lowercase.
      sourceName is the name of a file ...
      (a) containing only lowercase text and lowercase html markup, 
      (b) where the spans of interest start with spanStart and end with spanEnd, 
      (c) the spanStarts and spanEnds match up the way parentheses do, and 
      (d) there are no embedded spans.
      */
    public static String produceCounts(String sourceName, String target, boolean choice) {
        // initialize static fields for new run
        w= target;
        hits=0; 
        misses=0;
        text=(new In(sourceName)).readAll(); // String version of the data
        
        if (choice) {
            updateCounts();
            return "Hits: " + hits + "; Misses: " + misses + 
                (printRatioInfo && 
                 (double)hits/misses > unigramHitMissRatio?" * more hits than expected by chance":""); 
        } else {
            Pair p= hitsAndMisses(text, target);
            return "Hits: " + p.x + "; Misses: " + p.y + 
                (printRatioInfo && 
                  (double)hits/misses > unigramHitMissRatio?" * more hits than expected by chance":""); 
        }
    }
    
    /** update hits and misses as follows:
      *    hits incremented with the number of occurrences of w in spans in text
      *    misses incremented with the number of occurrences of w in non-spans in text.
      * 
      * Preconditions: (basically, same as for produceCounts())  
      * w consists only of letters, and at least one, all lowercase.
      * text is a String ...
      *   (a) containing only lowercase text and lowercase html markup, 
      *   (b) where the spans of interest start with spanStart and end with spanEnd, 
      *   (c) the spanStarts and spanEnds match up the way parentheses do, 
      *   (d) and there are no embedded spans.*/
    private static void updateCounts() {
        // System.out.println("  text is: " + text);  // DEBUG
        
        // Determine if w is in text
        int firstWIndex=text.indexOf(w);
        if (firstWIndex== -1) { 
            return; // no instances found, so no need to update
        }
        
        // w occurs in text
        
        /* if spanStart is before this first occurrence of w, 
         * and no spanEnd is between the last preceding spanStart and this w,
         * thenthe w is in a span. */
        int closestPrecedingSpanStartIndex=(text.substring(0,firstWIndex)).lastIndexOf(spanStart);  
        boolean inSpan= 
            closestPrecedingSpanStartIndex > -1  &&   
            (text.substring(closestPrecedingSpanStartIndex,firstWIndex)).indexOf(spanEnd) == -1;
        
        if (inSpan) {
            hits++;
        }
        else {
            misses++;
        }
        
        /* Process remaining instances of w in text after the one just processed.
         * If in a span, need to maintain the matching-parens invariant*/
        text= (inSpan?spanStart:"") + text.substring(firstWIndex+w.length());
        
        // System.out.println(text.substring(0,Math.min(20, text.length()))); //DEBUG
        
        updateCounts(); 
    }
    
    /** = index of first occurrence of s in text[i..] (or text.length() if none) */
    public static int next(String text, String s, int i) {
        i= text.indexOf(s, i);
        return (i >= 0 ? i : text.length());
    }
    
    /** = the number of hits and misses of word w in text t.
      For the Pair p returned, the numbers of hits and misses are in p.x nd p.y,
      repsectively. */
    public static Pair hitsAndMisses(String text, String word) {
        int istart= next(text, spanStart, 0);
        int iend= next(text, spanEnd, istart);
        int iword= next(text, word, 0);
        
        return hitsAndMisses(text, word, istart, iend, iword);
    }
    
    /** Return a pair (h, m) where h is the number of hits in text[iw..] and
      m is the number of misses in text[iw..].
      A hit is an occurrence of w that occurs within a span
      A miss is an occurrence of w that occurs outside a span.
      
      Precondition: text[iStart..iEnd] is a span (or both iStart and iEnd
      equal text.length()), and iw is either an index of word w in text or text.length().
      
      */
    private static Pair hitsAndMisses(String text, String w, int iStart, int iEnd, int iw) {
        if (iw == text.length())
            return new Pair(0, 0);
        
        if (iw < iStart) {
            // text[iw] is a miss
            Pair p= hitsAndMisses(text, w, iStart, iEnd, next(text, w, iw + w.length()));
            p.y= p.y + 1;
            return p;
        }
        
        /** text[iw..] is in or after the span. */
        if (iw < iEnd) {
            // text[iw] is a hit because it appears in the span
            Pair p= hitsAndMisses(text, w, iStart, iEnd, next(text, w, iw + w.length()));
            p.x= p.x + 1;
            return p; 
        }
        
        /** w occurs after the span, so this span is unnecessary */
        iStart= next(text, spanStart, iStart + spanStart.length());
        iEnd=  next(text, spanEnd, iStart);
        return hitsAndMisses(text, w, iStart, iEnd, iw);
    }
    
}